package dsql;

import dsql.exception.*;
import dsql.parser.*;
import dsql.value.*;

import java.sql.*;
import java.util.*;

/** Stores run-time metadata about a driver/database, and manages
 * new connections.
 * There is a corresponding instance of this class for each &lt;connect&gt;
 * element.  
 * The code is optimised to minimise the open connections
 * and calls to get metadata (if supported by the driver) */

public class Connect extends DMExecuted {

  String conid, driverId, driverClass, conurl;
  ASTUserName userName;
  ASTPassword password;

  Vector SQLconns = new Vector(); // Keeps track of connections
    // and the live statements. If the driver supports
    // multiple statements there will never be more than one entry here
    // (i.e. the same connection is reused for all statements),
    // otherwise each connection in this vector will be limited to
    // 0..1 statements.
    // each entry in the vector is: Vector(Connection,Vector(Statements*))*

  boolean supportsMultipleResultSets,
          checkResultSetType; // if true, make sure a result set is scrollable
                              // even if driver did not complain when setting
                              // the option
  boolean supportsConnMetaData = true; // sets to false only if getMetaData 
                              // fails (in which case, won't try it again
                              // for this connection); this is the case
                              // for example with unixODBC
  boolean supportsPrimaryKeys = true; // sets to false if cannot get primary
                              // keys on this connection
  boolean supportsUpdatableSets = false; // true if ResultSets can be updated
  boolean computeDisplaySize = false; // if true, do not trust getColDispSize
  boolean gotConnMetaData = false; // set to true once setMetaData has 
                              // been called
  boolean supportsMultipleStatements = false; // set to true if there can
                              // be more than one open statement per
                              // connection
  int     canReadRandomColumns = 1;  // -1: unknown; 0: no; 1: yes
  int     supportsScrollCursor = -1; // -1: unknown; 0: no; 1: yes

  public Connect(String _id, ASTDriver driver,
                 String _cString,
                 ASTUserName _userName, ASTPassword _password) {
    setCaller(driver.getCaller());
    conid = _id;
    driverId = driver.getId();
    driverClass = driver.getJavaClass();
    conurl = _cString;
    userName = _userName;
    password = _password;
    try {setDriverMetadata(driver);}
    catch (ExecuteException e) {
      System.err.println(e.toString());
      System.exit(-1);
    }
  }

  public String getId() {return conid;}
  public String getDriverId() {return driverId;}

  public Value execute(ASTSQL sqlObj,String sql) throws ExecuteException {
    return new SQLResults(sqlObj,sql);
  }

  public boolean supportsMultipleResultSets() {
    return supportsMultipleResultSets;
  }
  public boolean supportsMultipleStatements() {
    return supportsMultipleStatements;
  }
  public boolean checkResultSetType() {
    return checkResultSetType;
  }
  public boolean supportsConnMetaData() {
    return supportsConnMetaData;
  }
  public boolean supportsUpdatableSets() {
    return supportsUpdatableSets;
  }
  public boolean computeDisplaySize() {
    return computeDisplaySize;
  }
  public int canReadRandomColumns() {
    return canReadRandomColumns;
  }
  public int supportsScrollCursor() {
    return supportsScrollCursor;
  }
  public void setCanReadRandomColumns(int c) {
    canReadRandomColumns = c;
  }

  public void closeStatement() throws ExecuteException {
    throw new ExecuteException(
      "Method Connect.closeStatement() not implemented yet!");
  }

  public Statement openStatement() throws ExecuteException {
    ConnectStmt connStmt = getConnection();
    int concur = supportsUpdatableSets() ? ResultSet.CONCUR_UPDATABLE :
	ResultSet.CONCUR_READ_ONLY;
    if(supportsScrollCursor()!=0) // i.e. either unknown or true
    try {
      Statement stmt = connStmt.addStatement(
        ResultSet.TYPE_SCROLL_INSENSITIVE,concur);
      supportsScrollCursor = 1;
      return stmt;
    } catch (SQLException exc) {
      supportsScrollCursor = 0;
      caller.trace.warning("Connection "+getId()+" with driver "+getDriverId()+
        " does not support scrollable cursors. Please see "+
        "the driver's documentation for details and DSQL "+
        "documentation for performance issues.");
    }
    try {
      return connStmt.addStatement(ResultSet.TYPE_FORWARD_ONLY,concur);
    } catch (SQLException e) {
      throw new ExecuteException("Could not create statement: "+
        e.getMessage(),e);
    }
  }

  public String[] getPrimaryKeys(String catalog,String schema,String table) 
      throws ExecuteException {
    if(!supportsPrimaryKeys) return null;
    if(!supportsConnMetaData()) {
      supportsPrimaryKeys = false;
      throw new ExecuteException("Unabble to read primary keys: driver "+
        getDriverId()+" does not support metadata");
    }
    try {
      ResultSet keySet = getMetadata().getPrimaryKeys(catalog,schema,table);
      Vector keyV = new Vector();
      while(keySet.next()) {
        String colName = keySet.getString(4);
        keyV.addElement(colName.intern());
      }
      String[] arr = new String[keyV.size()];
      keyV.copyInto(arr);
      return arr;
    } catch (SQLException exc) {
      supportsPrimaryKeys = false;
      throw new ExecuteException("Unable to read primary keys: "+
                                 exc.getMessage());
    }
  }

  protected ConnectStmt getConnection() throws ExecuteException {
    if(SQLconns.isEmpty()) // if no connection yet, create one
      return openConnection();
    if(supportsMultipleStatements())
      // always return the same connection
      return (ConnectStmt) SQLconns.firstElement();
    else {
      // try to find a connection without statements
      for (Enumeration e = SQLconns.elements() ; e.hasMoreElements() ;) {
        ConnectStmt conStmt = (ConnectStmt) e.nextElement();
        if(!conStmt.hasStatements()) return conStmt;
      }
      // couldn't find one
      return openConnection();
    }
  }

  /** Add an entry in the connection vector and returns the connection object 
  */
  protected ConnectStmt openConnection() throws ExecuteException {
    try {
      if(driverClass!=null) Class.forName(driverClass);
      caller.trace.sql("Connecting to "+conurl);
      Connection con = 
        DriverManager.getConnection(conurl,userName.getUserName(),
        	                    password.getPassword());
      ConnectStmt conStmt = new ConnectStmt(con);
      SQLconns.addElement(conStmt);
      setMetadata();
      return conStmt;
    } catch (ClassNotFoundException exc) {
      throw new ExecuteException("Unable to use driver "+getDriverId()+
        ": class "+driverClass+" could not be found.");
    } catch (SQLException exc) {
      throw new ExecuteException("Unable to establish connection to '"+
        getId()+"': "+exc.getMessage(),exc);
    }
  }

  protected DatabaseMetaData getMetadata() throws ExecuteException {
    if(SQLconns.isEmpty()) throw new ExecuteException("Attempt to get "+
      "metadata before a connection was opened for "+getId());
    if(!supportsConnMetaData()) throw new ExecuteException("Driver "+
      getDriverId()+" does not support metadata");
    // get metadata from first connection
    try {return ((ConnectStmt) SQLconns.firstElement()).
      getConnection().getMetaData();
    } catch (SQLException exc) {
      throw new ExecuteException(exc.getMessage());
    }
  }

  /** Get connection metadata; will be called just once per Connect object */
  protected void setMetadata() throws ExecuteException {
    if(!supportsConnMetaData() || gotConnMetaData) return;
    gotConnMetaData = true;
    try {
      DatabaseMetaData metadata = getMetadata();
      try {supportsMultipleResultSets = metadata.supportsMultipleResultSets();}
      catch (SQLException e) {supportsMultipleResultSets=true;}
      try {if(metadata.supportsResultSetType(
             ResultSet.TYPE_SCROLL_INSENSITIVE)) supportsScrollCursor = 1;
      } catch (SQLException e) {} // still unknown!
      try {if((supportsScrollCursor==1 && 
       metadata.supportsResultSetConcurrency(ResultSet.TYPE_SCROLL_INSENSITIVE,
					     ResultSet.CONCUR_UPDATABLE)) ||
       metadata.supportsResultSetConcurrency(ResultSet.TYPE_FORWARD_ONLY,
					     ResultSet.CONCUR_UPDATABLE))
	supportsUpdatableSets = true;
      } catch (SQLException e) {}
    } catch (ExecuteException exc) {
      supportsMultipleResultSets = true;
      supportsConnMetaData = false;
      caller.trace.warning("Unable to get metadata for "+
          getId()+": "+exc.getMessage());
    }
  }

  protected void setDriverMetadata(ASTDriver driver) throws ExecuteException {
    if(driver.getElement("checkResultSetType")!=null)
      checkResultSetType = true;
    if(driver.getElement("supportMultipleStatements")!=null)
      supportsMultipleStatements = true;
    if(driver.getElement("computeDisplayColumnSize")!=null)
      computeDisplaySize = true;
    ASTXMLElement randomColumnsEl = driver.getElement("canReadRandomColumns");
    if(randomColumnsEl!=null) {
      String value = randomColumnsEl.getXMLText();
      if(value.equals("yes")) canReadRandomColumns = 1;
      else if(value.equals("no")) canReadRandomColumns = 0;
      else if(value.equals("check")) canReadRandomColumns = -1;
      else throw new ExecuteException(
        "invalid value for <canReadRandomColumns>: "+value+" (should be: "+
        "yes, no or check)");
    }
  }

/*
//-------------------------------------------------------------------
// checkForWarning
// Checks for and displays warnings.  Returns true if a warning
// existed
// invoke with:	checkForWarning (con.getWarnings ());
//-------------------------------------------------------------------
private static boolean checkForWarning (SQLWarning warn) 	
		throws SQLException  {
	boolean rc = false;

	// If a SQLWarning object was given, display the
	// warning messages.  Note that there could be
	// multiple warnings chained together

	if (warn != null) {
		System.out.println ("\n *** Warning ***\n");
		rc = true;
		while (warn != null) {
			System.out.println ("SQLState: " +
				warn.getSQLState ());
			System.out.println ("Message:  " +
				warn.getMessage ());
			System.out.println ("Vendor:   " +
				warn.getErrorCode ());
			System.out.println ("");
			warn = warn.getNextWarning ();
		}
	}
	return rc;
}
  */

}
