package dsql;

import dsql.exception.*;
import dsql.parser.*;

import java.io.*;

/*
 * DSQL.java
 *
 * Created on July 16, 2003, 12:43 PM
 */

/** The DSQL (Dynamic SQL) interpreter. It is invoked from the command-line
 * with the "-i"
 * switch to start the GUI, or with a file name to execute.
 * @author Francois Paradis
 */
public class DSQL extends Thread {

  protected static String traceMode;
  protected static boolean interactive = false;
  protected static boolean openedWindow = false;
  protected static boolean parseOnly = false;

  /**
   * @param args the command line arguments
   */
  public static void main(String args[]) {
    if(args.length==0) {
      System.err.println("* Reading script from standard input...");
      int r = new DSQL(new Parser(System.in),"standard input",
                                  args,new Trace()).execute();
      System.exit(r);
    } else {
      int i = 0;
      while(i<args.length)
        i = command(args,i);
      if(interactive && !openedWindow) new DSQLInterface();
    }
  }

  public static int command(String args[], int i) {
    try {
    String cmd = args[i];
    if(cmd.equals("-encrypt") || cmd.equals("-e")) {
      if(i+1<args.length) {
        encrypt(args[i+1]);
        return i+2;
      } else {
        System.err.println("* ERROR: Provide encryption method");
        System.exit(-1);
      }
    } else if(cmd.equals("-trace") || cmd.equals("-t")) {
      if(i+1<args.length) {
        traceMode = args[i+1];
        return i+2;
      } else {
        System.err.println("* ERROR: Provide trace mode");
        System.exit(-1);
      }
    } else if(cmd.equals("-parseonly")) {
      parseOnly = true;
    } else if(cmd.equals("-interactive") || cmd.equals("-i")) {
      interactive = true;
    } else {
      String[] newArgs = new String[args.length-i-1];
      for(int j=0; j<newArgs.length; j++)
	  newArgs[j] = args[j+i+1];
      if(interactive) {
	new DSQLInterface(args[i]);
        openedWindow = true;
      } else {
        int r = new DSQL(new Parser(ASTInclude.getFile(args[i])),args[i],
	  	         newArgs,new Trace(traceMode)).execute();
        System.exit(r);
      }
    }
    } catch (ExecuteException e) {
      System.err.println(e.toString());
      System.exit(-1);        
    }
    return i+1;
  }

  public static void encrypt(String method) {
    try {
      String plain = ASTPassword.inputPassword("Enter password:");
      String encrypt = ASTPassword.encrypt(plain,method);
      String decrypt = ASTPassword.decrypt(encrypt,method);
      if(plain.equals(decrypt))
        System.out.println("<password encrypt=\""+method+"\">"+encrypt+
                           "</password>");
        else
        System.err.println("* ERROR when checking the validity of the "+
                           "encrypted password");
    } catch (Exception exc) {
      System.err.println(exc.getMessage());
    }
  }


  Parser parser;
  public Trace trace;
  String streamName;
  String[] args;
  PrintStream out, err;
  DSQLInterface caller;
  boolean requestToStop;

  public DSQL(Parser _parser, String _streamName, String[] _args, 
              Trace _trace) {
      this(_parser,_streamName,_args,_trace,System.out,System.err);
  }

  public DSQL(Parser _parser, String _streamName, String[] _args, 
              Trace _trace, PrintStream _out, PrintStream _err) {
    parser = _parser;
    parser.setCaller(this);
    streamName = _streamName;
    args = _args;
    trace = _trace;
    out = _out;
    err = _err;
  }

  public void start(DSQLInterface _caller) {
    caller = _caller;
    start();
  }

  public void setRequestToStop() {requestToStop=true;}

  /** Standard Thread method: do not call! use start() instead */
  public void run() {execute();}

  public boolean hasRequestToStop() {return requestToStop;}

  public void clearRequestToStop() {
    requestToStop=false; 
    if(caller!=null) caller.setExecuting(null);
  }

  public int execute() {
    try {
      trace.debug("*** Parsing "+streamName+"...");
      if(hasRequestToStop()) return -1;
      ASTDoc doc = parser.Document(streamName);
      if(hasRequestToStop()) return -1;
      if(parseOnly) {
        if(trace.isdebug()) doc.dump("");
        return 0;
      }
      trace.debug("*** Executing...");
      if(args!=null) doc.setArgs(args);
      doc.setOut(out);
      if(hasRequestToStop()) return -1;
      doc.execute();
      return 0;
    } catch (TokenMgrError e) {
      return printParseError(e,null,err);
    } catch (ParseException e) {
      return printParseError(new ParseExceptionDsql(e),null,err);
    } catch (ExecuteException e) {
      if(e.getCause() instanceof ParseException)
        // when a parse/lexical exception occurs inside an included file,
        // it is encapsulated in an ExecuteException
        return printParseError(
          new ParseExceptionDsql((ParseException)e.getCause()), 
          e.getMessage(),err);
      else if(e.getCause() instanceof TokenMgrError)
        return printParseError(e.getCause(), e.getMessage(),err);
      else {
	if(e instanceof ExitException && ((ExitException)e).errorno()==0)
	  return 0;
        err.print("* Error");
        if(e.getNode()!=null) {
          if(e.getNode().getRoot().isIncluded())
            err.print(", file \""+
                             e.getNode().getRoot().getFileName()+"\"");
          if(e.getNode().getLine()!=-1)
            err.print(", line "+e.getNode().getLine());
        } 
        err.print(": ");
        err.println(e.getMessage());
        return -1;
      }
    } finally {
      clearRequestToStop();
    }
  }

  public static int printParseError(Throwable exc, String file, 
                                    PrintStream err) {
    err.print("* Syntax Error"+
      (file!=null?", file \""+file+"\"":"")+": ");
    err.println(exc.getMessage());
    return -1;
  }

}
