package dsql;

import dsql.exception.*;
import dsql.parser.*;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.undo.*;

/** The DSQL GUI.
 * @author Francois Paradis
 */
public class DSQLInterface implements ActionListener, ComponentListener,
				      CaretListener, UndoableEditListener,
                                      DocumentListener, WindowListener {

  static Properties properties = null;
  static DSQLInterface root = null;
  protected static final int NB_RECENT_FILES = 10;
  protected static final Vector recentVector = new Vector(); 

  String fileName;
  PrintStream outStream;
  String cmdLineArgs;

  DSQL executeThread;

  boolean dirty = false;

    JFrame frame; // this window
  JMenu recentMenu;
    JMenuItem newItem, openItem, saveItem, saveAsItem, closeItem;
    JMenuItem[] recentMenuItem;
    JMenuItem undoItem, redoItem, selectAll, gotoLineItem;
    JMenuItem insertConnect, insertDSQL, insertSQL, insertQuote;
    JRadioButtonMenuItem courierFont,helveticaFont,size10,size12,size14,
	noWarnings,noItem,yesItem,sqlItem,debugItem;
    JMenuItem cmdLineItem;
    JMenuItem aboutItem;
    JTextArea editArea, outText;
    JPanel editPane;
    JScrollPane outScroll;
    JButton executeButton, stopButton, clearButton;
    JLabel caretPos;

  final UndoManager undoManager = new UndoManager();

  public DSQLInterface() {
    this(null);
  }

  public DSQLInterface(String _fileName) {
    this(_fileName,-1,-1);
  }

  public DSQLInterface(String _fileName, int x, int y) {
    if(root==null) {
      root = this;
      loadProperties();
    }
    if(x==-1) x = getPropertyInt(PROP_POSX);
    if(y==-1) y = getPropertyInt(PROP_POSY);
    frame = createWindow(_fileName);
    frame.setLocation(x,y);
    int width = getPropertyInt(PROP_SIZEX),
      height = getPropertyInt(PROP_SIZEY);
    if(width>0 && height>0) 
      frame.setSize(width,height);
    frame.addComponentListener(this); // needs to be set after location & size!
    frame.setVisible(true);
  }

  protected static final String PROP_POSX = "dsql.pos.x";
  protected static final String PROP_POSY = "dsql.pos.y";
  protected static final String PROP_SIZEX = "dsql.size.x";
  protected static final String PROP_SIZEY = "dsql.size.y";
  protected static final String PROP_FONT = "dsql.font";
  protected static final String PROP_FONTSIZE = "dsql.font.size";
  protected static final String PROP_TRACE = "dsql.trace";
  protected static final String PROP_RECENT = "dsql.recent.file";
  protected static final String PROP_DIRECTORY = "dsql.current.directory";

  protected void loadProperties() {
    properties = new Properties();
    try {
      properties.load(new FileInputStream(getSystemDirectory("user.home")+
					  "/.dsql"));
    } catch (Exception exc) {
      // initialise
      properties.setProperty(PROP_POSX, "0");
      properties.setProperty(PROP_POSY, "0");
      properties.setProperty(PROP_SIZEX, "0");
      properties.setProperty(PROP_SIZEY, "0");
      properties.setProperty(PROP_FONT, "Courier");
      properties.setProperty(PROP_FONTSIZE, "12");
      properties.setProperty(PROP_TRACE, "No");
      properties.setProperty(PROP_DIRECTORY, getSystemDirectory("user.dir"));
      // Initialise recent files with doc/examples 
      String exampleDir = new File(getSystemDirectory("DSQL_HOME")).
           getAbsolutePath()+"/doc/examples/";
      String[] entries = new File(exampleDir).list();
        if(entries!=null)
	  for(int i=0,j=0; i<entries.length && j<10; i++)
	    if(entries[i].endsWith(".xml"))
              properties.setProperty(PROP_RECENT+String.valueOf(++j),
	  			     exampleDir+entries[i]);
    }
    // setup recent vector
    int i =1;
    String recentProp;
    while((recentProp=getProperty(PROP_RECENT+String.valueOf(i++)))!=null)
      recentVector.add(recentProp);
  }

  protected void saveProperties() {
    if(this==root)
    try {
      properties.store(new FileOutputStream(getSystemDirectory("user.home")
					    +"/.dsql"),
		       "DSQL Interface Settings");
    } catch (Exception exc) {
      System.err.println("* Unable to save settings: "+exc);
    }
  }

  protected int getPropertyInt(String key) {
    return Integer.parseInt(getProperty(key));
  }

  protected int getPropertyInt(String key, int _default) {
      String r=getProperty(key);
      if(r==null) return _default;
      else return Integer.parseInt(r);      
  }

  protected String getProperty(String key) {
    return properties.getProperty(key);
  }

  protected String getProperty(String key,String _default) {
    return properties.getProperty(key,_default);
  }

  protected void setProperty(String key, int value) {
    setProperty(key,String.valueOf(value));
  }

  protected void setProperty(String key, String value) {
    if(this==root)
      properties.setProperty(key,value);
  }

  protected String getSystemDirectory(String propName) {
    return System.getProperty(propName,".");
  }

  protected JFrame createWindow(String _fileName) {
    frame = new JFrame();
    frame.addWindowListener(this);
    setFileName(_fileName);
    frame.setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
    // Prepare edit pane
    editArea = createEditArea();
    JScrollPane editScroll = new JScrollPane(editArea);
    executeButton = new JButton("Execute");
    executeButton.setMnemonic(KeyEvent.VK_E);
    executeButton.setActionCommand("execute");
    executeButton.addActionListener(this);
    stopButton = new JButton("Stop");
    stopButton.setActionCommand("stop");
    stopButton.addActionListener(this);
    editPane = new JPanel();
    editScroll.setPreferredSize(new Dimension(500,200));
    editPane.setLayout(new BoxLayout(editPane, BoxLayout.PAGE_AXIS));
    editPane.add(editScroll);
    JPanel edButtonPane = new JPanel();
    edButtonPane.setLayout(new BoxLayout(edButtonPane, BoxLayout.LINE_AXIS));
    edButtonPane.add(Box.createHorizontalGlue());
    caretPos = new JLabel("1:1");
    edButtonPane.add(caretPos);
    edButtonPane.add(executeButton);
    edButtonPane.add(stopButton);
    editPane.add(edButtonPane);
    // Prepare output pane
    outText = new JTextArea();
    outText.setEditable(false);
    outScroll = new JScrollPane(outText);
    outScroll.setPreferredSize(new Dimension(500,100));
      //      JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
      //JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS);
    JTabbedPane tabPane = new JTabbedPane();
    tabPane.addTab("Output",outScroll);
    JPanel outPane = new JPanel();
    outPane.setLayout(new BoxLayout(outPane, BoxLayout.PAGE_AXIS));
    outPane.add(tabPane);
    JPanel outButtonPane = new JPanel();
    outButtonPane.setLayout(new BoxLayout(outButtonPane, BoxLayout.LINE_AXIS));
    outButtonPane.add(Box.createHorizontalGlue());
    clearButton = new JButton("Clear");
    clearButton.setActionCommand("clear");
    clearButton.addActionListener(this);
    outButtonPane.add(clearButton);
    outPane.add(outButtonPane);
    // Set menu
    frame.setJMenuBar(createMenuBar());
    // Set contents
    JSplitPane spane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
				      editPane,outPane);
    spane.setOpaque(true);
    frame.setContentPane(spane);
    editPane.getRootPane().setDefaultButton(executeButton);
    frame.pack();
    if(_fileName!=null) openFile(_fileName);
    setFont(getProperty(PROP_FONT,"Courier"),getPropertyInt(PROP_FONTSIZE,12));
    setExecuting(null);
    outStream = new PrintStream(new JTextAreaOutputStream(outText));
    System.setOut(outStream); // TBD: should only be done for first window?
    System.setErr(outStream);
    return frame;
  }


  protected JTextArea createEditArea() {
    JTextArea editArea = new JTextArea();
    createActionTable(editArea);
    Document doc = editArea.getDocument();
    // Listen for undo and redo events
    doc.addUndoableEditListener(this);
    doc.addDocumentListener(this);
    // Create an undo action and add it to the text component
    editArea.getActionMap().put("Undo",
        new AbstractAction("Undo") {
            public void actionPerformed(ActionEvent evt) {
                try {if (undoManager.canUndo()) {undoManager.undo();}
                } catch (CannotUndoException e) {}
            }
       });
    // Bind the undo action to CTRL-Z
    editArea.getInputMap().put(
      KeyStroke.getKeyStroke(KeyEvent.VK_Z, ActionEvent.CTRL_MASK), "Undo");
    // Create a redo action and add it to the text component
    editArea.getActionMap().put("Redo",
        new AbstractAction("Redo") {
            public void actionPerformed(ActionEvent evt) {
                try {if (undoManager.canRedo()) {undoManager.redo();}
                } catch (CannotRedoException e) {}
            }
        });
    // Bind the redo action to CTRL-Y
    editArea.getInputMap().put(
      KeyStroke.getKeyStroke(KeyEvent.VK_Y, ActionEvent.CTRL_MASK), "Redo");
    editArea.addCaretListener(this);
    return editArea;
  }


  protected JMenuBar createMenuBar() {
    JMenuBar menuBar = new JMenuBar();
    // ***** File Menu
    JMenu fileMenu = new JMenu("File");
    fileMenu.setMnemonic(KeyEvent.VK_F);
    newItem = createMenuItem(fileMenu,"New Window",KeyEvent.VK_N,
      KeyStroke.getKeyStroke(KeyEvent.VK_N, ActionEvent.CTRL_MASK));
    openItem = createMenuItem(fileMenu,"Open...",KeyEvent.VK_O,
      KeyStroke.getKeyStroke(KeyEvent.VK_O, ActionEvent.CTRL_MASK));
    saveItem = createMenuItem(fileMenu,"Save",KeyEvent.VK_S,
      KeyStroke.getKeyStroke(KeyEvent.VK_S, ActionEvent.CTRL_MASK));
    saveAsItem = createMenuItem(fileMenu,"Save as...");
    fileMenu.addSeparator();
    recentMenu = new JMenu("Recent");
    setupRecentMenu();
    fileMenu.add(recentMenu);
    fileMenu.addSeparator();
    closeItem = createMenuItem(fileMenu,"Close",KeyEvent.VK_C);
    fileMenu.add(closeItem);
    menuBar.add(fileMenu);
    // ***** Edit Menu
    JMenu editMenu = new JMenu("Edit");
    editMenu.setMnemonic(KeyEvent.VK_E);
    undoItem = createMenuItem(editMenu,"Undo");
    redoItem = createMenuItem(editMenu,"Redo");
    editMenu.addSeparator();
    selectAll = createMenuItem(editMenu,"Select All");
    editMenu.add(getActionByName(DefaultEditorKit.cutAction));
    editMenu.add(getActionByName(DefaultEditorKit.copyAction));
    editMenu.add(getActionByName(DefaultEditorKit.pasteAction));
    editMenu.addSeparator();
    gotoLineItem = createMenuItem(editMenu,"Goto line...",KeyEvent.VK_G,
      KeyStroke.getKeyStroke(KeyEvent.VK_G, ActionEvent.CTRL_MASK));
    menuBar.add(editMenu);
    // ***** Insert Menu
    JMenu insertMenu = new JMenu("Insert");
    insertMenu.setMnemonic(KeyEvent.VK_I);
    insertConnect = createMenuItem(insertMenu,"<connect>",0,
       KeyStroke.getKeyStroke(KeyEvent.VK_C, ActionEvent.ALT_MASK));
    insertDSQL = createMenuItem(insertMenu,"<dsql>",0,
       KeyStroke.getKeyStroke(KeyEvent.VK_D, ActionEvent.ALT_MASK));
    insertQuote = createMenuItem(insertMenu,"<quote>",0,
       KeyStroke.getKeyStroke(KeyEvent.VK_Q, ActionEvent.ALT_MASK));
    insertSQL = createMenuItem(insertMenu,"<sql>",0,
       KeyStroke.getKeyStroke(KeyEvent.VK_S, ActionEvent.ALT_MASK));
    menuBar.add(insertMenu);
    // ***** Options Menu
    JMenu optionsMenu = new JMenu("Options");
    optionsMenu.setMnemonic(KeyEvent.VK_O);
    JMenu fontMenu = new JMenu("Font");
    ButtonGroup fontGroup = new ButtonGroup();
    courierFont = createRadioButtonMenuItem(fontMenu,fontGroup,"Courier",
					    PROP_FONT);
    helveticaFont = createRadioButtonMenuItem(fontMenu,fontGroup,"Helvetica",
					      PROP_FONT);
    //    courierFont.setSelected(true);
    JMenu fontSizeMenu = new JMenu("Font Size");
    ButtonGroup fontSizeGroup = new ButtonGroup();
    size10 = createRadioButtonMenuItem(fontSizeMenu,fontSizeGroup,"10",
				       PROP_FONTSIZE);
    size12 = createRadioButtonMenuItem(fontSizeMenu,fontSizeGroup,"12",
				       PROP_FONTSIZE);
    size14 = createRadioButtonMenuItem(fontSizeMenu,fontSizeGroup,"14",
				       PROP_FONTSIZE);
    // size12.setSelected(true);
    JMenu traceMenu = new JMenu("Trace");
    ButtonGroup traceGroup = new ButtonGroup();
    noWarnings = createRadioButtonMenuItem(traceMenu,traceGroup,
					   "No warnings",PROP_TRACE);
    noItem = createRadioButtonMenuItem(traceMenu,traceGroup,"No",PROP_TRACE);
    // noItem.setSelected(true);
    yesItem = createRadioButtonMenuItem(traceMenu,traceGroup,"Yes",PROP_TRACE);
    sqlItem = createRadioButtonMenuItem(traceMenu,traceGroup,"SQL",PROP_TRACE);
    debugItem = createRadioButtonMenuItem(traceMenu,traceGroup,
					  "Debug",PROP_TRACE);
    optionsMenu.add(fontMenu);
    optionsMenu.add(fontSizeMenu);
    optionsMenu.add(traceMenu);
    cmdLineItem = createMenuItem(optionsMenu,"Command line...");
    menuBar.add(optionsMenu);
    // ***** Help Menu
    JMenu helpMenu = new JMenu("Help");
    aboutItem = createMenuItem(helpMenu,"About DSQL...");
    menuBar.add(helpMenu);
    return menuBar;
  }


  protected JMenuItem createMenuItem(JMenu parentMenu, String txt) {
    return createMenuItem(parentMenu,txt,0,null);
  }

  protected JMenuItem createMenuItem(JMenu parentMenu, String txt, 
				     int mnemonic) {
    return createMenuItem(parentMenu,txt,mnemonic,null);
  }

  protected JMenuItem createMenuItem(JMenu parentMenu, String txt, 
				     int mnemonic, KeyStroke accelerator) {
    JMenuItem mi = new JMenuItem(txt);
    if(mnemonic!=0) mi.setMnemonic(mnemonic);
    if(accelerator!=null) mi.setAccelerator(accelerator);
    mi.addActionListener(this);
    parentMenu.add(mi);
    return mi;
  }

  protected JRadioButtonMenuItem createRadioButtonMenuItem(JMenu parentMenu,
	       ButtonGroup group, String txt, String property) {
    JRadioButtonMenuItem button = new JRadioButtonMenuItem(txt);
    if(getProperty(property,"").equals(txt))
      button.setSelected(true);
    button.addActionListener(this);
    group.add(button);
    parentMenu.add(button);
    return button;
  }

  protected void setFileName(String _fileName) {
    fileName = _fileName;
    frame.setTitle((fileName!=null)?"DSQL: "+fileName:"DSQL");
  }

  protected void insertText(String txt) {
    if(editArea.getSelectedText()!=null)
      editArea.replaceSelection(txt);
    else
      editArea.insert(txt,editArea.getCaretPosition());
  }

  protected void insertTag(String tag) {
    insertTag(tag,null);
  }
  protected void insertTag(String tag, String attrs) {
    insertTag(tag,attrs,null);
  }
  protected void insertTag(String tag, String attrs, String contents) {
    insertText("<"+tag+(attrs!=null?" "+attrs:"")+">"+
	       (contents!=null?contents:"")+"</"+tag+">");
    editArea.setCaretPosition(editArea.getCaretPosition()-tag.length()-3);
  }

  protected JFileChooser createFileChooser() {
    JFileChooser fc = new JFileChooser(new File(getProperty(PROP_DIRECTORY,
							    ".")));
    // TBD: eventually there should be a filer on files? 
    return fc;
  }


  protected boolean unsavedEdits() {
    return dirty;
  }

  protected void openFile(String _fileName) {
    try {
      editArea.read(new FileReader(_fileName),_fileName);
      setFileName(_fileName);
      dirty = false;
      editArea.getDocument().addDocumentListener(this);
      editArea.getDocument().addUndoableEditListener(this);
      updateRecent();
    } catch (FileNotFoundException exc) {
      showError("File "+_fileName+" does not exist","I/O Error");
    } catch (IOException exc) {
      showError("Error while reading "+_fileName,"I/O Error");
    }
  }

  protected void saveFile(String _fileName) {
    try {
      editArea.write(new FileWriter(_fileName));
      setFileName(_fileName);
      dirty = false;
      editArea.getDocument().addDocumentListener(this);
      editArea.getDocument().addUndoableEditListener(this);
      updateRecent();
    } catch (IOException exc) {
      showError("Error while writing "+_fileName,"I/O Error");
    }
  }

  protected boolean saveDialog() {
    JFileChooser fc = createFileChooser();
    int rval = fc.showSaveDialog(editPane);
    updateCurrentDirectory(fc);
    if(rval == JFileChooser.APPROVE_OPTION) {
      File file = fc.getSelectedFile();
      String name = fc.getSelectedFile().getAbsolutePath();
      if(file.exists()) {
        if(!chooseYesNo("File "+file.getName()+" already exists.\n"+
			"Do you want to overwrite it?","File exists"))
          return false;
      }
      saveFile(file.getAbsolutePath());
      return true;
    } else
      return false;
  }


  protected void showError(String msg, String title) {
    JOptionPane.showMessageDialog(frame,msg,title,JOptionPane.ERROR_MESSAGE);
  }

  protected boolean chooseYesNo(String msg, String title) {
    return JOptionPane.showConfirmDialog(frame,msg,title,
      JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION;
  }

  protected int chooseYesNoCancel(String msg, String title) {
    int i = JOptionPane.showConfirmDialog(frame,msg,title,
      JOptionPane.YES_NO_CANCEL_OPTION);
    return i;
    //==JOptionPane.YES_OPTION;
  }

  protected void setFont(String font,int size) {
    editArea.setFont(new Font(font,Font.PLAIN,size));
    outText.setFont(new Font(font,Font.PLAIN,size));
    setProperty(PROP_FONT,font);
    setProperty(PROP_FONTSIZE,size);
    saveProperties();
  }

  protected void setTrace(String trace) {
      setProperty(PROP_TRACE,trace);
  }

  protected void setExecuting(DSQL _thread) {
    executeThread = _thread;
    boolean executing = _thread!=null;
    executeButton.setEnabled(!executing);
    stopButton.setEnabled(executing);
  }


  protected void updateSizePosition() {
    setProperty(PROP_POSX,frame.getX());
    setProperty(PROP_POSY,frame.getY());
    setProperty(PROP_SIZEX,frame.getWidth());
    setProperty(PROP_SIZEY,frame.getHeight());
  }

  protected void updateCurrentDirectory(JFileChooser fileChooser) {
    setProperty(PROP_DIRECTORY,
		fileChooser.getCurrentDirectory().getAbsolutePath());
    saveProperties();
  }

  protected void updateRecent() {
    int found;
    if((found=recentVector.indexOf(fileName))!=-1) {
      if(found==0) return;
      recentVector.removeElementAt(found);
    }
    recentVector.insertElementAt(fileName,0);   
    int size=recentVector.size();
    if(size>10) recentVector.setSize(10);
    for(int i=0; i<size; i++)
      setProperty(PROP_RECENT+String.valueOf(i+1),
		  recentVector.elementAt(i).toString());
    setupRecentMenu(); 
    saveProperties();
 }

  protected void setupRecentMenu() {
    recentMenu.removeAll();
    int size = recentVector.size();
    recentMenuItem = new JMenuItem[size];
    for (int i=0; i<size; i++)
      recentMenuItem[i] = createMenuItem(recentMenu,
	 new File(recentVector.elementAt(i).toString()).getName());
  }

  // TBD: should not be used anymore?
  // replace cut&paste with my own items

  HashMap actions;
  protected void createActionTable(JTextComponent textComponent) {
    actions = new HashMap();
    Action[] actionsArray = textComponent.getActions();
    for (int i = 0; i < actionsArray.length; i++) {
      Action a = actionsArray[i];
      actions.put(a.getValue(Action.NAME), a);
    }
  }

  protected Action getActionByName(String name) {
    return (Action)(actions.get(name));
  }

  /** checks if file needs to be saved and aborts if user
   * asks to save it and then cancels */
  protected boolean abortOpen() {
    if(unsavedEdits()) {
      // bring up save dialog
      int i=chooseYesNoCancel("Save changes to "+
			      (fileName==null?"buffer":fileName)+
			      "?","Unsaved");
      if(i==JOptionPane.CANCEL_OPTION) return true;
      if(i==JOptionPane.YES_OPTION) {
	if(fileName!=null) saveFile(fileName);
	else if(!saveDialog()) return true;
      }
    }
    return false;
  }

  public String[] getCmdLineArgs() {
    if(cmdLineArgs==null) return null;
    String[] args = new String[1];
    args[0] = cmdLineArgs;
    return args;
  }

  // ***********************************
  // LISTENERS
  // ***********************************

  public void actionPerformed(ActionEvent e) {
    JComponent source = (JComponent)(e.getSource());
    if(source==newItem) {
      new DSQLInterface(null,frame.getX()+10,frame.getY()+10); 
      return;
    }
    if(source==openItem) {
      if(abortOpen()) return;
      JFileChooser fc = createFileChooser();
      int rval = fc.showOpenDialog(editPane);
      updateCurrentDirectory(fc);
      if(rval == JFileChooser.APPROVE_OPTION) {
        openFile(fc.getSelectedFile().getAbsolutePath());
        return;
      }
    }
    if(source==saveItem) {
      if(fileName!=null) saveFile(fileName);
      else saveDialog();
      return;
    }
    if(source==saveAsItem) {
      saveDialog();
      return;
    }
    if(source==closeItem) {
      frame.dispatchEvent(new WindowEvent(frame,
        WindowEvent.WINDOW_CLOSING));
    }
    if(source==undoItem) {
      if(undoManager.canUndo()) undoManager.undo();
      return;
    }
    if(source==redoItem) {
      if(undoManager.canRedo()) undoManager.redo();
      return;
    }
    if(source==selectAll) {
      editArea.selectAll();
      return;
    }
    if(source==insertConnect) {
      insertTag("connect","id=\"default\" dbmsid=\"\"",
                "\n  <dbname></dbname>\n  <username></username>\n");
      return;
    }
    if(source==insertSQL) {
      insertTag("sql","connectid=\"default\"");
      return;
    }
    if(source==insertQuote) {
      insertTag("quote");
      return;
    }
    if(source==insertDSQL) {
      insertTag("dsql");
      return;
    }
    if(source==courierFont) {
	setFont("Courier",getPropertyInt(PROP_FONTSIZE,12));
      return;
    }
    if(source==helveticaFont) {
      setFont("Helvetica",getPropertyInt(PROP_FONTSIZE,12));
      return;
    }
    if(source==size10) {
      setFont(getProperty(PROP_FONT,"Courier"),10);
      return;
    }
    if(source==size10) {
      setFont(getProperty(PROP_FONT,"Courier"),10);
      return;
    }
    if(source==size12) {
      setFont(getProperty(PROP_FONT,"Courier"),12);
      return;
    }
    if(source==size14) {
      setFont(getProperty(PROP_FONT,"Courier"),14);
      return;
    }
    if(source==noWarnings) {
	setTrace("No warning");
	return;
    }
    if(source==noItem) {
	setTrace("No");
	return;
    }
    if(source==sqlItem) {
	setTrace("SQL");
	return;
    }
    if(source==yesItem) {
	setTrace("Yes");
	return;
    }
    if(source==debugItem) {
	setTrace("Debug");
	return;
    }
    if(source==executeButton) {
      String txt = editArea.getText();
      outText.setText("");
      if(txt.length()>0) {
	DSQL dsql = null;
	try {
          dsql = new DSQL(new Parser(new StringReader(txt)),fileName,
			  getCmdLineArgs(),
			  new Trace(outStream,getProperty(PROP_TRACE,"no")),
			  outStream,outStream);
	} catch (ExecuteException exc) {
	    outStream.println(exc);
        }
        setExecuting(dsql);
        dsql.start(this);
      }
      return;
    }
    if(source==stopButton) {
      if(executeThread!=null) executeThread.setRequestToStop();
      return;
    }
    if(source==clearButton) {
      outText.setText("");
      return;
    }
    for(int i=0; i<recentMenuItem.length; i++) {
      if(source==recentMenuItem[i]) {
        if(abortOpen()) return;
        openFile(recentVector.elementAt(i).toString());
        return;
      }
    }
    if(source==cmdLineItem) {
      String value = (String) JOptionPane.showInputDialog(
                    frame,
                    "Enter command-line arguments\n"+
                    "separated by spaces:","Command-line arguments",
                    JOptionPane.PLAIN_MESSAGE,
                    null,null,cmdLineArgs);
      if(value!=null) cmdLineArgs = value;
      return;
    }
    if(source==gotoLineItem) {
      String value = (String) JOptionPane.showInputDialog(
                    frame,"Line number:","Goto line",
                    JOptionPane.PLAIN_MESSAGE,null,null,"");
      if(value!=null)
        try {
	  int line = Integer.parseInt(value)-1; 
          int lineCount = editArea.getLineCount();
          if(line<=0) {editArea.setCaretPosition(0); return;}
          else if(line>=lineCount) line=lineCount-1;
          int ofs = editArea.getLineEndOffset(line-1);
          editArea.setCaretPosition(ofs);
	} catch (Exception exc) {}
      return;
    }
  }


  public void caretUpdate(CaretEvent ev) {
    int ofs = ev.getDot();
    try {
      int line = editArea.getLineOfOffset(ofs);
      caretPos.setText(""+(line+1)+":"+
       (ofs-editArea.getLineStartOffset(line)+1));
    } catch (BadLocationException exc) {}
  }


  public void undoableEditHappened(UndoableEditEvent evt) {
    undoManager.addEdit(evt.getEdit());
  }

  public void componentHidden(ComponentEvent e) {}
  public void componentMoved(ComponentEvent e) {updateSizePosition();}
  public void componentResized(ComponentEvent e) {updateSizePosition();}
  public void componentShown(ComponentEvent e) {}

  public void insertUpdate(DocumentEvent e) {dirty=true;}
  public void removeUpdate(DocumentEvent e) {dirty=true;}
  public void changedUpdate(DocumentEvent e) {}

  public void windowActivated(WindowEvent e) {}
  public void windowClosed(WindowEvent e) {}
  public void windowClosing(WindowEvent e) {
    if(abortOpen()) return;
    saveProperties();
    frame.dispose();
  }
  public void windowDeactivated(WindowEvent e) {}
  public void windowDeiconified(WindowEvent e) {}
  public void windowIconified(WindowEvent e) {}
  public void windowOpened(WindowEvent e) {}
 

}

