package dsql.parser;

import dsql.exception.*;
import dsql.value.*;
import dsql.tools.MaskingThread;

import java.io.*;
import java.math.*;

/** A <code>&lt;password&gt;</code> element */
public class ASTPassword extends ASTXMLElement {

  String method = "plain";

  public ASTPassword(int id) {
    super(id);
  }

  public ASTPassword(Parser p, int id) {
    super(p, id);
  }

  /** Constructor for ASTXMLElement.recast() */
  public ASTPassword(ASTXMLElement toCopy) {
    super(toCopy);
  }

  public void setAttribute(String attrName, String attrVal) {
    if(attrName.equals("encrypt"))
      method = attrVal;
    else
      super.setAttribute(attrName,attrVal);
  }

  public void setMethod(String _method) {
    method = _method;
  }

  public String toString() {
    return "<password encrypt=\""+method+"\">";
  }

  public String getPassword() throws ExecuteException {
    String pwd = getXMLText();
    if(isEncrypted())
      pwd = decrypt(pwd,method);
    return pwd;
  }

  public boolean isEncrypted() {
    return !method.equals("plain");
  }

  /**
   * AUTHOR: Qusay H. Mahmoud
   * http://java.sun.com/features/2002/09/pword_mask.html
   *@param prompt The prompt to display to the user.
   *@return The password as entered by the user.
   */

  public static String inputPassword(String prompt) throws IOException {
    // TBD: use GUI if available

    // password holder
    String password = "";
    MaskingThread maskingthread = new MaskingThread(prompt);
    Thread thread = new Thread(maskingthread);
    thread.start();
    // block until enter is pressed
    while (true) {
      char c = (char)System.in.read();
      // assume enter pressed, stop masking
      maskingthread.stopMasking();
      if (c == '\r') {
        c = (char)System.in.read();
        if (c == '\n') {
          break;
        } else {
          continue;
        }
      } else if (c == '\n') {
        break;
      } else {
        // store the password
        password += c;
      }
    }
    return password;
  }

  public static String encrypt(String password, String method) 
      throws ExecuteException {
    if(method.equals("dsql"))
      return encryptDSQL(password);
    else
      throw new ExecuteException("Unknown encryption method "+method);
  }

  public static String decrypt(String password, String method) 
      throws ExecuteException {
    if(method.equals("dsql"))
      return decryptDSQL(password);
    else
      throw new ExecuteException("Unknown encryption method "+method);
  }

  protected static String encryptDSQL(String plainText) 
      throws ExecuteException {
    int ptl = plainText.length();
    if(ptl+1>=pwdlength)
      throw new ExecuteException("Password is too long!");
    plainText = plainText + (char) (ptl*3);
    StringBuffer pwdstr = new StringBuffer(pwdlength);
    //    pwdstr.append((char)ptl);
    for(int i=0;i<(pwdlength-1-ptl);i++)
      pwdstr.append(plainText.charAt(((i+3)*29)%(ptl+1)));
    pwdstr.append(plainText);
    return encodeHex(xorString(pwdstr.toString()));
  }

  protected static String decryptDSQL(String hexText) 
      throws ExecuteException {
    String plainText = xorString(decodeHex(hexText));
    int ptl = ((int) plainText.charAt(pwdlength-1))/3;
    return plainText.substring(pwdlength-ptl-1,pwdlength-1);
  }

  protected static String xorString(String src) throws ExecuteException {
    int keylength = key.length();
    char[] dest = new char[src.length()];
    for(int i=0; i<src.length(); i++) {
      dest[i] = (char)(src.charAt(i) ^ key.charAt(i%keylength));
    }
    return new String(dest);
  }

  protected static String encodeHex(String in) throws ExecuteException {
    StringBuffer out = new StringBuffer();
    for(int i=0; i<in.length(); i++) {
      int cval = (int) in.charAt(i);
      if(cval>256) throw new ExecuteException("Cannot serialise non-ascii "+
                                              "character '"+in.charAt(i)+"'");
      out.append((cval<16?"0":"")+Integer.toHexString(cval));
    }
    return out.toString().toUpperCase();
  }

  protected static String decodeHex(String in) throws ExecuteException {
    int l = in.length();
    if(l%2!=0) throw new ExecuteException("Invalid length ("+l+
                                          ") for serialised password "+in);
    char[] out = new char[l/2];
    for(int i=0; i<out.length; i++)
      try {
        out[i] = (char)Integer.decode("0x"+in.substring(i*2,i*2+2)).intValue();
      } catch (NumberFormatException exc) {
        throw new ExecuteException(exc.getMessage());
      }
    return new String(out);
  }

  private static final int pwdlength = 16;
  private static final String key = 
    "le silence vertebral indispose la voile licite";

}
