package dsql.parser;

import dsql.DSQL;
import dsql.value.*;
import dsql.exception.*;

import java.lang.reflect.*;

/** A node element; this is the superclass of all parsed elements */

public abstract class DMNode extends SimpleNode {

  int line = -1;
  DSQL caller;

  public DMNode(int id) {
    super(id);
  }

  public DMNode(Parser p, int id) {
    super(p, id);
  }

  public DMNode(DMNode toCopy) {
    super(toCopy.parser,toCopy.id);
    parent = toCopy.parent;
    children = toCopy.children;
    setLine(toCopy.getLine());
    setCaller(toCopy.getCaller());
  }

  public DSQL getCaller() {return caller;}
  public void setCaller(DSQL _caller) {caller=_caller;}

  public void jjtInsertFirstChild(Node n) {
    jjtAddChild(n,children.length);
    if(children.length>1)
      System.arraycopy(children, 0, children, 1, children.length-1);
    children[0] = n;
  }

  public void jjtSetChild(int i,Node n) {
    children[i] = n;
  }

  public ASTDoc getRoot() {
    DMNode n = this;
    while(n.jjtGetParent()!=null)
      n = (DMNode) n.jjtGetParent();
    return (ASTDoc) n;
  }

  public ASTDsql getDsqlParent() {
    DMNode n = this;
    while((n=(DMNode)n.jjtGetParent())!=null && !(n instanceof ASTDsql)) {}
    return (ASTDsql) n;
  }

  public void setLine(int _line) {line=_line;}
  public int getLine() {return line;}

  /** Execute a Node
   * @throws ExecuteException If an error occurs during execution
   */
  public abstract Value execute() throws ExecuteException;

  /** Execute a Node
   * @throws ExecuteException If an error occurs during execution,
   * the exception is set with this Node so that a line number
   * can be displayed along with the error message
   */
  public Value executeWithCatch() throws ExecuteException {
    if(caller!=null && caller.hasRequestToStop()) 
      throw new StopException();
    try {return execute();}
    catch (ExecuteException exc) {if(exc.getNode()==null) exc.setNode(this); 
      throw exc;}
  }

  protected ASTXMLElement executeXML() throws ExecuteException {
    Value val = executeWithCatch();
    if(val instanceof XmlDataValue) 
      return ((XmlDataValue)val).getXmlData(true);
    else
      return new ASTXMLText(val.getString());
  }

  public boolean hasArgument(int i) {
    return i<jjtGetNumChildren();
  }
  public DMNode getArgument(int i) throws ExecuteException {
    if(hasArgument(i))
      return (DMNode) children[i];
    else 
      throw new ExecuteException("There is no argument #"+i);
  }
  public Value executeArgument(int i) throws ExecuteException {
    return getArgument(i).executeWithCatch();
  }
  /** Execute argument and ensure it returns a value */
  public Value executeArgumentNonNull(int i) throws ExecuteException {
    Value r = executeArgument(i);
    if(r==null) throw new ExecuteException(
      "Expression does not return a value");
    else return r;
  }

  /** Execute argument and ensure it is an instance of class cl 
   * @param tryToCast if true, try to convert the argument to class cl */
  public Object executeArgumentAsClass(int i,Class cl,boolean tryToCast) 
      throws ExecuteException {
    Value r = executeArgumentNonNull(i);
    if(!(cl.isInstance(r))) {
      if(tryToCast) return r.getType(cl);
      else throw new ExecuteException(r+" is not a "+cl);
    }
    else return r;
  }

  public TableResult executeArgumentTable(int i) throws ExecuteException {
    return (TableResult) executeArgumentAsClass(i,TableResult.class,false);
  }
  public ColumnHolder executeArgumentColumnHolder(int i) 
      throws ExecuteException {
    return (ColumnHolder) executeArgumentAsClass(i,ColumnHolder.class,false);
  }
  public RecordHolder executeArgumentRecordHolder(int i) 
      throws ExecuteException {
    return (RecordHolder) executeArgumentAsClass(i,RecordHolder.class,true);
  }

  public String getString() {return toString();}

}
