package dsql.value;

import dsql.DSQL;
import dsql.parser.*;
import dsql.exception.*;


import java.io.*;
import java.util.*;
import java.text.*;
import org.apache.poi.poifs.filesystem.POIFSFileSystem;
import org.apache.poi.hssf.usermodel.*;

/** Read an excel file using the POI library 
 * (http://jakarta.apache.org/poi/).
 * The results is the set of all sheets, each sheet correspond to a table */

public class ExcelDataValue extends AbstractResults implements DataValue {

  static final int STRING_CELL = 0;
  static final int NUMINT_CELL = 1;
  static final int NUMERIC_CELL = 2;
  static final int DATE_CELL = 3;
  static final int NUMPERC_CELL = 4;

  static String[] NATIVE_TYPES = {"STRING","NUMERIC","NUMERIC",
                                  "NUMERIC","NUMERIC"};
  static String[] SQL_TYPES = {"VARCHAR","INTEGER","FLOAT","DATE","DECIMAL"};

  HSSFWorkbook wb;
  HSSFDataFormat dataFormat;
  HSSFSheet curSheet;
  HSSFRow curRow;
  int curSheetIndex = 0, 
      curRecordIndex = 0; // current record (1-based)
  int firstRowIndex = 0, // 1st row with data (0-based)
      lastRowIndex = 0;

  String[] sheetLabels;
  int[] sheetTypes;
  String[] sheetFormats; // types format string
  String[] sheetInternalFormats;

  public ExcelDataValue() {}

  public String toString() {return "Excel file";}

  /** @param all if true, iterates through all results */
  public ConstructedResults getMeta(boolean all) throws ExecuteException {
    if(all) throw new ExecuteException("Meta all not implemented for "+
	        		       "ExcelValue");
    ConstructedResults r = new ConstructedResults(getCaller());
    addCurrentResultMeta(r);
    return r;
  }

  public void addCurrentResultMeta(ConstructedResults r) 
        throws ExecuteException {
      r.addTable(METALABELS);
      int n = getColumnCount();
      for(int i=1; i<=n; i++) {
	String[] colValues = new String[8];
	colValues[0] = getColumnLabel(i);
	colValues[1] = SQL_TYPES[sheetTypes[i-1]];
	colValues[2] = NATIVE_TYPES[sheetTypes[i-1]] +
	    (sheetFormats[i-1]==null ? "":":"+sheetFormats[i-1]);
	colValues[3] = String.valueOf(getColumnDisplaySize(i));
	colValues[4] = "?"; // precision
	colValues[5] = "?"; // scale
        colValues[6] = "?";
        colValues[7] = "?";
	r.addRecord(colValues);
      }
  }

  // **************** DataValue interface

  public void setContents(ASTXMLElement ele) throws ExecuteException {
    throw new ExecuteException("ExcelDataValue do not allow inline values");
  }

  public void setContents(String _fileName, ASTXMLElement ele)
      throws ExecuteException {
    setCaller(ele.getCaller());
    try {
      POIFSFileSystem fs = new POIFSFileSystem(new FileInputStream(_fileName));
      wb = new HSSFWorkbook(fs);
      dataFormat = wb.createDataFormat();
      allreset();
    } catch (FileNotFoundException e) {
      throw new ExecuteException("File not found: \""+_fileName+"\"");
    } catch (IOException exc) {
      throw new ExecuteException("Error reading excel file "+_fileName);
    }
  }


  // ******************* AbstractResults methods
  //                     each sheet is a table result

  public void allreset() throws ExecuteException {
    gotoSheet(0);
  }

  public boolean hasResult() {return curSheet!=null;}
  public boolean nextResult() throws ExecuteException {
    curSheetIndex++;
    if(curSheetIndex<wb.getNumberOfSheets()) {
      gotoSheet(curSheetIndex);
      return true;
    } else {
      curSheet = null;
      curRow = null;
      curRecordIndex = 0;
      return false;
    }
  }

  // Table methods

  public void reset() {
    gotoRecord(0);
  }

  public String[] readCurrentRecord() throws ExecuteException {
    int colCount = sheetLabels.length;
    String[] values = new String[colCount];
    for(int i=0; i<colCount; i++)
      values[i] = getStringCellValue((short)i);
    return values;
  }

  public int getRecordCount() {
    return lastRowIndex-firstRowIndex+1;
  }

  public boolean nextRecord() {
   gotoRecord(currentRecord()+1);
   return curRow!=null;
  }

  /** first record is 1 */
  public void gotoRecord(int pos) {
    curRecordIndex = pos;
    int rowNb = curRecordIndex-1+firstRowIndex;
    getCaller().trace.trace("Moving to record #"+pos+" (row #"+rowNb+")");
    if(rowNb<firstRowIndex || rowNb>lastRowIndex)
      curRow = null;
    else
      curRow = curSheet.getRow(rowNb);
  }

  public int currentRecord() {
    return curRecordIndex;
  }

  // Tabular methods

  /** First column is 1 */
  public String getColumnString(int i) throws ExecuteException {
    if(i<1 || i>sheetLabels.length)
      throw new ExecuteException("Cell #"+i+" is out of range");
    if(curRow==null) return "";
    getCaller().trace.trace("Getting column #"+i);
    return getStringCellValue((short)(i-1));
    // getCellType: CELL_TYPE_STRING, CELL_TYPE_NUMERIC, CELL_TYPE_FORMULA, CELL_TYPE_BOOLEAN, CELL_TYPE_ERROR
    //    return cell.getNumericCellValue();
  }
  public Object getColumn(int i) throws ExecuteException {
    return getColumnString(i);
  }

  public String getColumnString(String s)  throws ExecuteException {
    return getColumnString(findColumn(s));
  }
  public Object getColumn(String s) throws ExecuteException {
    return getColumnString(s);
  }

  public int getColumnCount() throws ExecuteException {
    if(sheetLabels!=null) return sheetLabels.length;
    else return 0;
  }

  public String getColumnLabel(int i) {
    return sheetLabels[i-1];
  }

  // TBD: calculate!
  public int getColumnDisplaySize(int i) throws ExecuteException {
    return 20; // default column size
  }

  //  .....

  public void gotoSheet(int index) throws ExecuteException {
    curSheetIndex = index;
    getCaller().trace.trace("Moving to sheet "+curSheetIndex);
    curSheet = wb.getSheetAt(curSheetIndex);
    firstRowIndex = curSheet.getFirstRowNum();
    lastRowIndex = curSheet.getLastRowNum();
    sheetLabels = getSheetLabels();
    sheetTypes = getSheetTypes();
    sheetFormats = getSheetFormats();
    sheetInternalFormats = getSheetInternalFormats();
    reset();
  }


  public void insert(ColumnHolder rec) throws ExecuteException {
    throw new ExecuteException("Excel sheet modification not supported");
  }
  public void deleteRecord() throws ExecuteException{
    throw new ExecuteException("Excel sheet modification not supported");
  }
  public void update(ColumnHolder rec) throws ExecuteException {
    throw new ExecuteException("Excel sheet modification not supported");
  }

  protected int findColumn(String label) throws ExecuteException {
    if(sheetLabels!=null)
      for(int i=0; i<sheetLabels.length; i++)
        if(sheetLabels[i].equals(label)) return i;
    throw new ExecuteException("No column "+label);
  }


  protected String getStringCellValue(short i) {
    HSSFCell cell = curRow.getCell(i);
    if(cell==null) return null;
    switch (sheetTypes[i]) {
      case STRING_CELL: 
        return cell.getStringCellValue();
      case NUMINT_CELL: 
        try {return String.valueOf((int)cell.getNumericCellValue());}
        catch (NumberFormatException exc) {return null;}
      case NUMPERC_CELL:
      case NUMERIC_CELL:
        try {return String.valueOf((float)cell.getNumericCellValue());}
	catch (NumberFormatException exc) {return null;}
      case DATE_CELL:
        SimpleDateFormat formatter = 
          new SimpleDateFormat(sheetInternalFormats[i]);
        Date dateVal;
        try {dateVal = cell.getDateCellValue();}
        catch (NumberFormatException exc) {return null;}
	if(dateVal==null) return null;
        return formatter.format(dateVal);
      default: return "";
    }
  }


  protected int getCellType(HSSFCell cell) {
    switch (cell.getCellType()) {
      case HSSFCell.CELL_TYPE_NUMERIC: 
        short formatIndex = cell.getCellStyle().getDataFormat();
        String format = dataFormat.getFormat(formatIndex);
        switch (formatIndex) {
  	  case 0: return NUMERIC_CELL;
 	  case 1: case 3: case 5: case 6: 
                  return NUMINT_CELL;
	  case 0xe:  case 0xf:  case 0x10: case 0x11:
	  case 0x12: case 0x13: case 0x14: case 0x15:
	  case 0x16: case 0x2d:	case 0x2e: case 0x2f:
                  return DATE_CELL;
	}
        if(format.indexOf('0')==-1)
	  return DATE_CELL;
        if(format.indexOf('%')!=-1) 
	  return NUMPERC_CELL;
        if(format.indexOf('.')==-1) 
	  return NUMINT_CELL;
        return NUMERIC_CELL;
      case HSSFCell.CELL_TYPE_STRING: return STRING_CELL;
      default: return STRING_CELL;
        
    }
  }

  protected String convertDateFormatString(String formatStr ) {
    // after first H, 'm' means minute
    int firstH = formatStr.indexOf('h');
    if(firstH==-1) formatStr = formatStr.replace('m','M');
    else formatStr = formatStr.substring(0,firstH).replace('m','M')
	   +formatStr.substring(firstH);
    if(formatStr.endsWith("AM/PM"))
      formatStr = formatStr.substring(0,formatStr.length()-5)+"a";
    if(formatStr.endsWith(";@"))
      formatStr = formatStr.substring(0,formatStr.length()-2);
    int backslash = -1;
    while((backslash=formatStr.indexOf('\\',backslash+1))!=-1) {
      char c = formatStr.charAt(backslash);
      if(c=='-' || c=='\\' || c=='/') 
	  formatStr=formatStr.substring(0,backslash)+
                    formatStr.substring(backslash+1);
    }
    return formatStr;
  }


  /** find labels from first non-empty row, null if not found
   * side-effect: set firstRowIndex */
  protected String[] getSheetLabels() throws ExecuteException {
    getCaller().trace.trace("Getting sheet labels...");
    curRow = null;
    while(curRow==null && firstRowIndex<=lastRowIndex)
      curRow = curSheet.getRow(firstRowIndex++);
    getCaller().trace.debug("Set data first row to: "+firstRowIndex);
    if(curRow==null) return null;
    Vector v = new Vector();
    int lastCol = curRow.getLastCellNum();
    for(short i=0; i<=lastCol; i++) {
      HSSFCell cell = curRow.getCell(i);
      if(cell!=null && cell.getCellType()!=HSSFCell.CELL_TYPE_STRING)
	  throw new ExecuteException("Invalid value for label at column "+
				     (i+1)+", line "+firstRowIndex);
      v.add(cell==null ? "" : cell.getStringCellValue());
    }
    while(((String)v.lastElement()).length()==0)
      v.remove(v.size()-1);
    String[] labels = new String[v.size()];
    v.copyInto(labels);
    return labels;
  }

  /** find types from first non-empty row, null if not found
   * must be called after getSheetLabels()
   * side-effect: set firstRowIndex */
  protected int[] getSheetTypes() throws ExecuteException {
    if(sheetLabels==null) return null;
    getCaller().trace.trace("Getting sheet types...");
    curRow = curSheet.getRow(firstRowIndex);
    while(curRow==null && firstRowIndex<=lastRowIndex)
      curRow = curSheet.getRow(++firstRowIndex);
    if(curRow==null) return null;
    int[] types = new int[sheetLabels.length];
    for(short i=0; i<types.length; i++)
      types[i] = getCellType(curRow.getCell(i));;
    return types;
  }

  protected String[] getSheetFormats() throws ExecuteException {
    if(sheetTypes==null || curRow==null) return null;
    getCaller().trace.trace("Getting sheet format strings...");
    String[] formats = new String[sheetTypes.length];
    for(short i=0; i<formats.length; i++) {
      short formatIndex = curRow.getCell(i).getCellStyle().getDataFormat();
      String format = dataFormat.getFormat(formatIndex);
      if(format.equals("General")) {
	if(sheetTypes[i]==NUMERIC_CELL && first100AreInteger(i))
	  sheetTypes[i] = NUMINT_CELL;
      } else
  	formats[i] = format;
    }
    return formats;
  }

  protected String[] getSheetInternalFormats() throws ExecuteException {
    if(sheetFormats==null || curRow==null) return null;
    getCaller().trace.trace("Getting sheet internal format strings...");
    String[] formats = new String[sheetFormats.length];
    for(short i=0; i<formats.length; i++) {
      if(sheetTypes[i]==DATE_CELL && sheetFormats[i]!=null)
        formats[i] = convertDateFormatString(sheetFormats[i]);
    }
    return formats;
  }

  /** there is no way to know if numeric cells with generic format
   * are reals or integers... this method returns true if the first
   * 100 rows do not include reals (or dates) */ 
  protected boolean first100AreInteger(short col) {
    int i = firstRowIndex;
    HSSFRow row;
    while((row=curSheet.getRow(i++))!=null && i<100) {
      HSSFCell cell = row.getCell(col);
      if(cell!=null) {
	if(cell.getCellType()==HSSFCell.CELL_TYPE_NUMERIC)
        try {
          float val = (float)cell.getNumericCellValue();
          if(val!=(float)((int)val)) return false;
	} catch (NumberFormatException exc) {return false;}
      }
    }
    return true;
  }

}
