package dsql.value;

import dsql.DSQL;
import dsql.Connect;
import dsql.exception.*;
import dsql.parser.ASTSQL;

import java.util.*;
import java.sql.*;

/** Results from an SQL query (i.e. wrapper around sql.ResultSet) */
public class SQLResults extends AbstractResults {

  Connect conn;
  ASTSQL sqlObj;       // the <sql>,<table> etc. element that generated
                       // this results
  Statement stmt = null; // the SQL statement associated to this result 
  String sql;          // the SQL statement string that produced this result,
                       // note: it could be re-executed if there is an attempt
                       // to go backwards on non-scrollable cursors
  ResultSet rs;        // current result set, or null
  int updateCount;     // current update count, or -1
  ResultSetMetaData rsmd = null;
  int recordIndex = 0; // current row index: 1 is first row
  int afterLastRecordIndex = -1; // index after last row if known, -1 otherwise
  int resultIndex = 1; // current result index: 1 is first
  //Vector savedResults = new Vector();

  int canReadRandomColumns = 1; // 1: can read columns in any order
                                // 0: can't
                                // -1: don't know
  boolean computeDisplaySize = false; // if true, don't trust getDispColSize

  Hashtable cachedLabels = null; // key=label; value=index
  String[] cachedIndex = null;   // index=index; value=label
  Vector cachedValues = null;    // index=index; value=value

  public SQLResults(ASTSQL _sqlObj, String _sql) 
      throws ExecuteException {
    sqlObj = _sqlObj;
    setCaller(sqlObj.getCaller());
    conn = _sqlObj.getConnect();
    canReadRandomColumns = conn.canReadRandomColumns();
    computeDisplaySize = conn.computeDisplaySize();
    sql = _sql;
    stmt = conn.openStatement();
    setResult(execute());
  }

  public String toString() {
    if(isResultSet())
      try {return "SQL Results "+rs.getCursorName();}
      catch (SQLException exc) {return "SQL Results";}
    else return "SQL Results";
  }

  public boolean hasResult() {
    return isResultSet() || isUpdate();
  }

  public void allreset() throws ExecuteException {
      // TBD: throw exception???
  }

  /** @return true if there was a result */
  public boolean nextResult() throws ExecuteException {
    if(!conn.supportsMultipleResultSets()) return false;
    if(!hasResult()) throw NoCurrentResultException();
    try {
      //      saveResult();
      boolean nr = setResult(stmt.getMoreResults());
      resultIndex++;
      if(nr) getCaller().trace.sql("Moved "+toString()+" to next result");
      return nr;
    } catch (SQLException exc) {throw ReadResultException(exc);}
  }

  public void reset() throws ExecuteException {
    gotoRecord(0);
  }

  public int currentRecord() throws ExecuteException {
    if(hasResult()) return recordIndex;
    else throw NoCurrentResultException();
  }

  public String[] readCurrentRecord() throws ExecuteException {
    int n = getColumnCount();
    String[] result = new String[n];
    if(canReadRandomColumns<1 && cachedValues==null) {
      readColumnsToCache();
      if(canReadRandomColumns==-1) checkRandomColumns();
    }
    if(cachedValues!=null) {
      growCachedColumns(n);
      cachedValues.copyInto(result);
    } else {
      for(int i=1; i<=n; i++)
        result[i-1] = getColumnString(i);
    }
    return result;
  }

  public void gotoRecord(int pos) throws ExecuteException {
    if(hasResult()) {
      if(pos==recordIndex || (pos==-1 && recordIndex==afterLastRecordIndex)) 
        return; // already at this record
      getCaller().trace.sql("Moving "+toString()+
                (pos==-1?" to last record":(pos==0?" before first record":
                  " to record #"+pos)));
      if(isResultSet()) {
        resultSetAbsolute(pos);
        clearCacheValues();
      } else if(isUpdate()) {
        if(pos<=2) recordIndex = pos;
        else throw GotoRecordException(pos);
      }
    } else throw NoCurrentResultException();
  }

  public int getRecordCount() throws ExecuteException {
    if(afterLastRecordIndex==-1) gotoRecord(-1);
    return afterLastRecordIndex-1;
  }

  /** @return true if there was a record, false otherwise */
  public boolean nextRecord() throws ExecuteException {
    boolean hasRecord;
    if(isResultSet())
      try {hasRecord = rs.next(); recordIndex++; clearCacheValues(); }
      catch (SQLException exc) {throw ReadResultException(exc);}
    else if(isUpdate()) {recordIndex += (recordIndex<=1?1:0);
                         hasRecord = recordIndex<=1;}
         else throw NoCurrentResultException();
    if(!hasRecord && afterLastRecordIndex==-1) 
      afterLastRecordIndex = recordIndex;
    getCaller().trace.sql("Moved "+toString()+
              (hasRecord?" to record #"+recordIndex:" after last record"));
    return hasRecord;
  }

  public String getColumnString(int i) throws ExecuteException {
    if(isResultSet())
      try { 
        if(canReadRandomColumns<1 && cachedValues==null) {
          readColumnsToCache();
          if(canReadRandomColumns==-1) checkRandomColumns();
        }
        if(cachedValues!=null) {
          growCachedColumns(i);
          return (String) cachedValues.elementAt(i-1);
        } else 
	  return rs.getString(i);
	//return rs.getObject(i); 
      } catch (SQLException exc) {throw ReadResultException(exc);}
    else if(isUpdate()) return String.valueOf(getUpdateCount());
    else throw NoCurrentResultException();
  }

  public Object getColumn(int i) throws ExecuteException {
    return getColumnString(i);
  }

  public String getColumnString(String s) throws ExecuteException {
    if(isResultSet())
      try { 
        if(canReadRandomColumns<1 && cachedValues==null) {
          readColumnsToCache();
          if(canReadRandomColumns==-1) checkRandomColumns();
        }
        if(cachedValues!=null) {
          Integer i = (Integer) cachedLabels.get(s);
          if(i!=null) {
            growCachedColumns(i.intValue());
            return (String) cachedValues.elementAt(i.intValue()-1);
          } else 
            throw new ExecuteException("no column "+s+" in results");
        } else
          return rs.getString(s); 
      } catch (SQLException exc) {throw ReadResultException(exc);}
    else if(isUpdate()) return String.valueOf(getUpdateCount());
    else throw NoCurrentResultException();
  }
  public Object getColumn(String s) throws ExecuteException {
    return getColumnString(s);
  }

  public int getColumnCount() throws ExecuteException {
    if(isResultSet())
      try { 
        if(cachedIndex!=null) return cachedIndex.length;
        else return getMetaData().getColumnCount();
      } catch (SQLException exc) { throw ReadResultException(exc); }
    else if(isUpdate()) return 1;
    else throw NoCurrentResultException();
  }


  public String getColumnLabel(int i) throws ExecuteException {
    if(isResultSet()) {
      // labels are cached...
      if(cachedIndex==null || cachedIndex[i-1]==null)
        // create/fill the cache if it does not exist yet
        try {
          if(cachedIndex==null) {
            int size = getMetaData().getColumnCount();
            cachedIndex = new String[size];
            cachedLabels = new Hashtable(size+1);
          }
          cachedIndex[i-1] = getMetaData().getColumnLabel(i);
          cachedLabels.put(cachedIndex[i-1],new Integer(i));
        } catch (SQLException exc) { throw ReadResultException(exc); }
      return cachedIndex[i-1];
    } else if(isUpdate()) return "RowsUpdated";
    else throw NoCurrentResultException();
  }


  public int getColumnDisplaySize(int i) throws ExecuteException {
    if(isResultSet()) {
      if(computeDisplaySize) return super.getColumnDisplaySize(i);
      else
      try {return getMetaData().getColumnDisplaySize(i);}
      catch (SQLException exc) {throw new ExecuteException(
        "Unable to get display size for column: ",exc);}
    } else if(isUpdate()) return super.getColumnDisplaySize(i);
    else throw NoCurrentResultException();
  }


  public int[] getPrimaryKeys() throws ExecuteException {
    if(super.getPrimaryKeys()!=null) return super.getPrimaryKeys();
    return setPrimaryKeys(sqlObj.getPrimaryKeys());
  }

  public ConstructedResults getMeta(boolean all) throws ExecuteException {
    ConstructedResults r = new ConstructedResults(getCaller());
    boolean more = hasResult();
    while(more) {
      addCurrentResultMeta(r);
      if(all) more = nextResult();
      else more = false;
    }
    return r; 
  }


  public void addCurrentResultMeta(ConstructedResults r) 
      throws ExecuteException {
    r.addTable(METALABELS);
    int n = getColumnCount();
    int[] keys = getPrimaryKeys();
    int keysIndex = 0;
    for(int i=1; i<=n; i++) {
      String[] colValues = new String[8];
      colValues[0] = getColumnLabel(i);
      if(isResultSet())
        try {
          ResultSetMetaData m = getMetaData();
          colValues[1] = getSQLTypeString(m.getColumnType(i));
          colValues[2] = m.getColumnTypeName(i);
          if(colValues[2]==null || colValues[2].equals("")) 
            colValues[2] = "?";
          colValues[3] = String.valueOf(m.getColumnDisplaySize(i));
          colValues[4] = String.valueOf(m.getPrecision(i));
          colValues[5] = String.valueOf(m.getScale(i));
          colValues[6] = String.valueOf(m.isNullable(i));
          if(primaryKeys==null) colValues[7] = "?";
          else {
            if(keysIndex<keys.length && keys[keysIndex]==i) {
              colValues[7] = "Y";
              keysIndex++;
            }
            else colValues[7] = "N";
          }
        } catch (SQLException exc) {
          throw new ExecuteException("Unable to get metadata: "+
                                     exc.getMessage(),exc);
        }
      else { // isUpdate
        colValues[1] = "INTEGER";
        colValues[2] = "INTEGER";
        colValues[3] = "11";
        colValues[4] = "?";
        colValues[5] = "?";
        colValues[6] = "0";
        colValues[7] = "Y";
      }
      r.addRecord(colValues);
    }
  }

  public void insert(ColumnHolder rec) throws ExecuteException {
    if(rs==null) 
      throw new ExecuteException("Unable to insert: no current result");
    if(conn.supportsUpdatableSets()) {
      try {
        rs.moveToInsertRow();
      } catch (SQLException exc) {    
	  throw new ExecuteException("Unable to insert into "+toString()+":"+
				     exc.getMessage(),exc);}
    } else {
      // result set cannot be updated: try sending an insert statement
    }
  }

  public void deleteRecord() throws ExecuteException{
    if(rs==null) 
      throw new ExecuteException("Unable to delete: no current result");
    if(conn.supportsUpdatableSets()) {
      try {
        rs.deleteRow();
      } catch (SQLException exc) {    
	  throw new ExecuteException("Unable to delete into "+toString()+":"+
				     exc.getMessage(),exc);}
    } else {
      // result set cannot be updated: try sending a delete statement
      throw new ExecuteException("Delete operation not yet implemented!");
    }
  }

  public void update(ColumnHolder rec) throws ExecuteException {
      // TBD
      /*    if(rs==null) return;
    try {rs.updateString(val.toString())}
    catch (SQLException exc) {    
      throw new ExecuteException("Unable to update "+toString()+
      ", col "+col+": "+exc.getMessage(),exc);}*/
  }


  // ********** PROTECTED *************

  /** needed by TableResult.setPrimaryKeys */
  protected int findColumn(String label) throws ExecuteException {
    if(cachedLabels!=null && cachedLabels.get(label)!=null)
      return ((Integer)cachedLabels.get(label)).intValue();
    else {
      if(rs==null) // not sure this can ever happen 
        throw new ExecuteException("No SQL results");
      try {return rs.findColumn(label);}
      catch (SQLException exc) {throw new ExecuteException(
        "Column "+label+" does not exist in results");}
    }
  }


  protected String getSQLTypeString(int type) {
    switch(type) {
      case Types.ARRAY  : return "ARRAY";
      case Types.BIGINT : return "BIGINT";
      case Types.BINARY : return "BINARY";
      case Types.BIT    : return "BIT";
      case Types.BLOB   : return "BLOB";
      case Types.CHAR   : return "CHAR";
      case Types.CLOB   : return "CLOB";
      case Types.DATE   : return "DATE";
      case Types.DECIMAL : return "DECIMAL";
      case Types.DISTINCT : return "DISTINCT";
      case Types.DOUBLE : return "DOUBLE";
      case Types.FLOAT : return "FLOAT";
      case Types.INTEGER : return "INTEGER";
      case Types.JAVA_OBJECT : return "JAVA_OBJECT";
      case Types.LONGVARBINARY : return "LONGVARBINARY";
      case Types.LONGVARCHAR : return "LONGVARCHAR";
      case Types.NULL : return "NULL";
      case Types.NUMERIC : return "NUMERIC";
      case Types.OTHER : return "OTHER";
      case Types.REAL : return "REAL";
      case Types.REF : return "REF";
      case Types.SMALLINT : return "SMALLINT";
      case Types.STRUCT : return "STRUCT";
      case Types.TIME : return "TIME";
      case Types.TIMESTAMP : return "TIMESTAMP";
      case Types.TINYINT : return "TINYINT";
      case Types.VARBINARY : return "VARBINARY";
      case Types.VARCHAR : return "VARCHAR";
      default : return "UNKNOWN";
    }
  }

  protected boolean isResultSet() {return rs!=null;}
  protected int getUpdateCount() {return updateCount;}
  protected boolean isUpdate() {return getUpdateCount()!=-1;}

  protected boolean setResult(boolean isResultSet) 
      throws ExecuteException {
    try {
      if(isResultSet) {
        rs = stmt.getResultSet(); 
        rsmd = null;
        updateCount = -1;
        afterLastRecordIndex = -1; // i.e. unknown
      } else {
        rs = null; 
        rsmd = null;
        updateCount = stmt.getUpdateCount();
        afterLastRecordIndex = 2; // only one record (the update count)
      }
      recordIndex = 0;
    } catch (SQLException exc) {
      throw ReadResultException(exc);
    }
    return hasResult();
  }

  /** goto the specified record of the current result set.
   */
  protected void resultSetAbsolute(int pos) throws ExecuteException {
    boolean forwardOnly = conn.supportsScrollCursor()!=1;
    if(!forwardOnly && conn.checkResultSetType())
      try {if(rs.getType()==ResultSet.TYPE_FORWARD_ONLY) {
        getCaller().trace.warning("The result set of connection "+conn.getId()+
                      " is not scrollable");
        forwardOnly = true;}
      } catch (SQLException exc) {}
    if(forwardOnly) { // this connection can only move
                      // forward one record at a time
      if(pos>=0 && pos<recordIndex) { // need to move backward: resend query
        int savedLastIndex = afterLastRecordIndex;
        setResult(execute());
        afterLastRecordIndex = savedLastIndex;
      }
      boolean readLast= (afterLastRecordIndex!=-1 &&
        recordIndex>=afterLastRecordIndex);
      int i = 0;
      while((recordIndex<pos || pos==-1) && !readLast)
        try {
          if(!rs.next()) readLast=true; 
          i++; recordIndex++;
        } catch (SQLException exc) {throw GotoRecordException(pos,exc);}
      if(readLast) afterLastRecordIndex = recordIndex;
      if(getCaller().trace.issql())
        getCaller().trace.sql("Moved "+toString()+
          " "+i+" positions (currently "+
          (!readLast?"at record #"+recordIndex+")":"after last record)"));
    } else try { // supports scrollable cursor
      if(pos==0) {rs.beforeFirst(); recordIndex = 0;}
      else
        if(pos==-1) {
          // don't use afterlast() because some drivers (e.g. MSSQL)
          // then set the row to -1
          rs.last(); 
          recordIndex=rs.getRow();
          rs.next();
          if(recordIndex!=-1) recordIndex++;
          afterLastRecordIndex=recordIndex;
        } else {if(rs.absolute(pos)) recordIndex = pos;
              else throw GotoRecordException(pos);}
    } catch (SQLException exc) {throw GotoRecordException(pos,exc);}
  }

  protected boolean execute() throws ExecuteException {
    getCaller().trace.sql("Sending: "+getCaller().trace.maxstr(sql,256));
    try {
      return stmt.execute(sql);
    } catch (SQLException exc) {
      throw new ExecuteException("Could not execute \""+getCaller().trace.maxstr(sql)+
        "\": "+exc.getMessage());
    }
  }

  /** Checks whether driver supports reading columns in random order
   * and set canReadRandomColumns accordingly.  */
  protected void checkRandomColumns() throws ExecuteException {
    if(canReadRandomColumns>=0) 
      return; // no need to check
    int nc = getColumnCount();
    if(nc<=1) {
      canReadRandomColumns=1; // 1 column (or less!): no need to cache
                              // this result anymore (but the connection
                              // object should stay to unknown)
      return;
    }
    getCaller().trace.trace("Checking if driver supports reading random columns");
    try { rs.getString(1); canReadRandomColumns = 1; }
    catch (SQLException exc) { 
      getCaller().trace.warning("Driver "+conn.getDriverId()+
                    " does not support reading random columns!");
      canReadRandomColumns = 0; 
    }
    // set the connection object, so from now on, results from this
    // driver won't need to be checked anymore
    conn.setCanReadRandomColumns(canReadRandomColumns);
  }


  protected void readColumnsToCache() throws ExecuteException {
    int nc = getColumnCount();
    cachedIndex = new String[nc];
    cachedLabels = new Hashtable(nc+1);
    cachedValues = new Vector(nc);
    growCachedColumns(nc);
  }


  protected void growCachedColumns(int i) throws ExecuteException {
    if(i>cachedValues.size())
      for(int j=cachedValues.size()+1; j<=i; j++) 
        try {
          if(cachedIndex[j-1]==null) {
            // cache labels if not already done!
            cachedIndex[j-1] = getMetaData().getColumnLabel(j);
            cachedLabels.put(cachedIndex[j-1],new Integer(j));
          }
          cachedValues.add(rs.getString(j));
          getCaller().trace.debug("Cached column "+cachedIndex[j-1]);
        } catch (SQLException exc ) {throw ReadResultException(exc);}
  }


  protected void clearCacheValues() {
    if(cachedValues!=null)
      if(canReadRandomColumns==1)  // do not cache from now on!
        cachedValues=null;
      else 
        cachedValues.clear();
  }


  /*
  protected void saveResult() throws ExecuteException {
    Object savedResult;
    if(isResultSet()) savedResult = rs;
    else if(isUpdate()) savedResult = new Integer(getUpdateCount());
    else throw NoCurrentResultException();
    savedResults.add(savedResult);
  }
  */

  protected ResultSetMetaData getMetaData() throws ExecuteException {
    if(rsmd==null) {
      try {rsmd = rs.getMetaData(); return rsmd;}
      catch (SQLException exc) {throw ReadResultException(exc);}
    } else
      return rsmd;
  }
  
  protected ExecuteException ReadResultException(SQLException exc) {
    return new ExecuteException("Unable to read "+toString()+": "+
                                exc.getMessage(),exc);
  }

  protected ExecuteException GotoRecordException(int pos) {
    return new ExecuteException("Unable to move to record "+
                                String.valueOf(pos));
  }

  protected ExecuteException GotoRecordException(int pos, SQLException exc) {
    return new ExecuteException("Unable to move to record "+
                                String.valueOf(pos)+": "+
                                exc.getMessage(),exc);
  }

  protected ExecuteException NoCurrentResultException() {
    return new ExecuteException("No current result to read from");
  }


}
