package dsql.value;

import dsql.*;
import dsql.exception.*;
import dsql.parser.ASTPrint;

import java.lang.reflect.*;

/** A value, ie the result of a DSQL expression. */
public abstract class Value extends DMExecuted {

  /** @return a string representation of this value, to be used
   * within dsql expressions. This is distinct from toString(),
   * which is used to output the object for debugging purposes. */
  public abstract String getString() throws ExecuteException;

  /** @return true if this Value is "non-zero", e.g. non-empty string,
   * tables has some record, etc. */
  public abstract boolean getBoolean() throws ExecuteException;

  /** @return this Value as an integer */
  public int getInt() throws ExecuteException {
    try {return Integer.parseInt(getString());}
    catch (NumberFormatException exc) {
      throw new ExecuteException("Expression "+getString()+" is not a "+
                                 "valid integer",exc);
    }
  }

  /** @return this Value as a float number */
  public float getFloat() throws ExecuteException {
    try {return Float.parseFloat(getString());}
    catch (NumberFormatException exc) {
      throw new ExecuteException("Expression "+getString()+" is not a "+
                                 "valid float",exc);
    }
  }

  /** {@link dsql.value.XmlDataValue XmlDataValue} uses this
   * to provide an iterator for XML documents */
  public RecordHolder getRecordHolder() throws ExecuteException {
      throw new ExecuteException("Value "+this.getClass()+" cannot be cast "+
				  "to RecordHolder");
  }

  public Object getType(Class cl) throws ExecuteException {
    if(cl==String.class) return getString();
    if(cl==Boolean.class) return new Boolean(getBoolean());
    if(cl==Integer.class) return new Integer(getInt());
    if(cl==Float.class) return new Float(getFloat());
    if(cl==RecordHolder.class) return getRecordHolder();
    throw new ExecuteException("Cannot convert "+this.toString()+
			       " to "+cl);
  }

  /** Attempt to return a Java object, e.g. StringValue returns String,
   * JavaObject returns the embedded object, etc. This is used when
   * passing values to java constructors and methods */
  public Object getObject() {return this;}

  /** return the object class; some scalars return the 
   * primitive class (eg Integer.TYPE, rather than Integer.getClass() */
  public Class getObjectClass() {
    return getClass();
  }

  /** Access an object's field */
  public Object getField(String name) throws ExecuteException {
    try {
      return getValue(getCaller(),getClass().getField(name).get(this));
    } catch (NoSuchFieldException exc) {
      String available;
      Field[] fields = getClass().getFields();
      if(fields.length>0) {
        StringBuffer f = new StringBuffer(fields[0].getName());
        for(int i=1; i<fields.length; i++) f.append(","+fields[i].getName());
        available = "Valid fields are: "+f.toString()+"."; 
      } else available = "No fields are available for this object.";
      throw new ExecuteException("Field '"+name+"' does not exist. "
                                 +available);
    } catch (IllegalAccessException exc) {
      throw new ExecuteException("Field '"+name+"' cannot be accessed",exc);
    }
  }

  public void setField(String name,Value val) throws ExecuteException {
    throw new ExecuteException("Cannot assign a value to "+name);
  }

  /** @return true if this value is a scalar, e.g. number, string, etc. */
  public abstract boolean isScalar();

  // ----------------------------------------

  /** Cast an arbitrary object to a Value.
   * @return the same object if it is already a Value */
  public static Value getValue(DSQL _caller,Object obj) {
    return getValue(_caller,obj,null);
  }

  /** This method is used by ASTMethodCall because of primitive return
   * types; it will cast an IntValue for a primitive int, but
   * will make a JavaObject for an Integer object
   */
  public static Value getValue(DSQL _caller,Object obj,Class cl) {
    if(obj instanceof Value)
      return (Value) obj;
    else
      if(obj==null) return new NullValue(_caller);
      else if(cl==Boolean.TYPE) 
        return new BooleanValue(_caller,(Boolean)obj);
      else if(cl==Integer.TYPE) 
        return new IntValue(_caller,(Integer)obj);
      else if(cl==Float.TYPE) 
        return new FloatValue(_caller,(Float)obj);
      else if(obj instanceof String) 
        return new StringValue(_caller,(String)obj);
      else return JavaObject.create(_caller,obj);
  }

  // TBD: should be in a subclass rather than here?
  // Structure For Meta Information table

  static final String[] METALABELS = { "ColumnName", "SQLType", "NativeType",
                           "DisplaySize",
                           "Precision", "Scale", "Nullable", "PrimaryKey" };

  /** @return meta-information about this Value, e.g. for a table,
   * it returns the names,types,width,etc. of its columns. */
  public abstract ConstructedResults getMeta() throws ExecuteException;

  /** Print this value using printer */
  public abstract void print(ASTPrint printer) throws ExecuteException;
}
