package dsql.value;

import dsql.DSQL;
import dsql.exception.*;
import dsql.parser.*;

import java.io.*;

/** Access to XML structures. External files are read progressively
 * and "on demand", i.e. whenever a "column" is requested the file
 * is read until a matching element is found. */

public class XmlDataValue extends NonScalarValue 
                          implements DotAt, DataValue {

  /** The original &lt;data&gt; element. */
  protected ASTXMLElement dataEle; 

  /** A buffer to some or all XML elements */
  protected ASTXMLElement buffer;  // XML data

  /** If true, read and parse more columns from external file.
   * If false, the whole structure must be in buffer. */
  protected boolean readColumnFromFile;

  /** the parser for external file */
  protected Parser parser;
  /** the external file name, or "inline" */
  protected String fileName;
  /** true if data is read from an external file */
  protected boolean external;


  /** constructor invoked by ASTValue */
  public XmlDataValue() {
  }

  public XmlDataValue(ASTXMLElement _ele) throws ExecuteException {
    setContents(_ele);
  }

  /** keeps track of record holders, so that we don't
   * reparse documents if user writes $x[0], $x[1] */
  protected XmlRecordHolder recordHolder = null;
  public RecordHolder getRecordHolder() throws ExecuteException {
    if(recordHolder==null) recordHolder = new XmlRecordHolder(this);
    return recordHolder;
  }

  /** Parse next element.
   * @return the element read, null otherwise
   */
  protected ASTXMLElement parseNext() throws ExecuteException {
    try {
      boolean allowDSQL = allowDSQL();
      ASTXMLElement next = parser.XMLNextElement(allowDSQL,allowDSQL);
      if(next!=null) {
        String tagName = next.getTagName();
        getCaller().trace.debug("Parsed "+(tagName==null?"XML text":
					   "<"+tagName+">"));
        return next;
      } else {
	// no more elements
        getCaller().trace.debug("No more contents to parse");
	return null;
      }
    } catch (ParseException exc) {
      throw new ExecuteException(fileName,exc);
    } catch (TokenMgrError exc) {
      throw new ExecuteException(fileName,exc);
    }
  }

  /** Parse next column and add it to buffer */
  protected ASTXMLElement parseNextColumn() throws ExecuteException {
    ASTXMLElement next = parseNext();
    if(next!=null)
      buffer.jjtAddChild(next,buffer.jjtGetNumChildren());
    else
      readColumnFromFile = false;
    return next;
  }

  /** Initialise parser and loads document header */
  protected void initParser() throws ExecuteException {
    if(external && parser==null) {
      getCaller().trace.debug("*** Parsing "+fileName+" header...");
      FileReader fileReader;
      // TBD: set encoding
      try {fileReader = new FileReader(fileName);}
      catch (FileNotFoundException exc) {
        throw new ExecuteException("File not found: "+fileName);}
      parser = new Parser(fileReader);
      parser.setCaller(getCaller());
      // parse header
      try {
        boolean allowDSQL = allowDSQL();
        buffer = parser.XMLHeader(allowDSQL,allowDSQL);
	// TBD: detect empty document, i.e. <doc/>
      } catch (ParseException exc) {
	  throw new ExecuteException(fileName,exc);
      } catch (TokenMgrError exc) {
	  throw new ExecuteException(fileName,exc);
      }
    }
  }

  protected ASTXMLElement getXmlData() throws ExecuteException  {
    initParser();
    return buffer;
  }

  public ASTXMLElement getXmlData(boolean full) 
      throws ExecuteException  {
    // TBD: if full, make sure the entire structure is read in
    return getXmlData();
  }

  public boolean equals(Object obj) {
    if(obj instanceof ScalarValue)
      try {
        return getString().equals(((ScalarValue)obj).getString());
      } catch (ExecuteException exc) {}
    return false;
  }

    /** @return true if DSQL scripts and variables are allowed
     * inside this XML; usually true for inline and false
     * for external */
  protected boolean allowDSQL() {
      return !external;
  }

  // **************** DataValue methods

  /** Inline */
  public void setContents(ASTXMLElement ele) throws ExecuteException {
    setCaller(ele.getCaller());
    fileName = "inline";
    external = false;
    readColumnFromFile = false;
    dataEle = ele;
    buffer = ele;
  }

  /** This method sets up but do not open the file for parsing.
   * This is delayed until the first real access */
  public void setContents(String _fileName, ASTXMLElement ele)
      throws ExecuteException {
    setCaller(ele.getCaller());
    fileName = _fileName;
    external = true;
    readColumnFromFile = true;
    dataEle = ele;
  }

  // ********************** Value methods

  static final String[] XMLTYPELABELS = { "SQLType", "NativeType", "File" };
  public ConstructedResults getMeta() throws ExecuteException {
      ConstructedResults r = new ConstructedResults(getCaller());
    r.addTable(XMLTYPELABELS);
    String[] colValues = new String[3];
    colValues[0] = "?";
    colValues[1] = "XML";
    colValues[2] = fileName;
    r.addRecord(colValues);
    return r;
  }

  public boolean getBoolean() throws ExecuteException {
    return getXmlData().jjtGetNumChildren()>0;
  }

  public String getString() throws ExecuteException {
    return getXmlData().getXMLText();
  }

  public void print(ASTPrint printer) throws ExecuteException {
      // TBD: loop to print everything!
    getXmlData().print(printer);
  }

  /** Access an element */
  public Object getField(String name) throws ExecuteException {
    return getColumn(name);
  }
  /** Access an element */
  public Object getField(int i) throws ExecuteException {
    return getColumn(i);
  }
  public Object setField(int i,Value val) throws ExecuteException {
    throw new ExecuteException("XmlDataValue.setField(int,Value) not implemented yet");
  }

  // ********************** ColumnHolder methods
    // no longer needed!

  public String getColumnLabel(int col) throws ExecuteException {
    return ((ASTXMLElement)getXmlData().jjtGetChild(getColumnIndex(col)))
      .getTagName();
  }

  public int getColumnCount() throws ExecuteException {
    throw new ExecuteException("XmlDataValue.getColumnCount() not implemented");
  }

  public Object getColumn(int col) throws ExecuteException {
    return getXmlData().jjtGetChild(getColumnIndex(col));
  }

  public String getColumnString(int col) throws ExecuteException {
    return ((Value)getColumn(col)).getString();
  }

  public Object getColumn(String name) throws ExecuteException {
    ASTXMLElement found = getXmlData().getElement(name);
    if(found!=null)
      return new XmlDataValue(found);
    else if(readColumnFromFile) {
      // not found, try to read more elements
      while((found = parseNextColumn())!=null) {
        String eleName = found.getTagName();
	if(eleName!=null && found.getTagName().equals(name)) 
          return new XmlDataValue(found);
      }
    }
    throw new ExecuteException("Element '"+name+"' does not exist");
  }

  public String getColumnString(String name) throws ExecuteException {
      return ((Value)getColumn(name)).getString();
  }

  // *******************************************

  protected int getColumnIndex(int col) throws ExecuteException {
    if(col<1) 
      throw new ExecuteException("Column "+col+" is out of bounds (<1)");
    ASTXMLElement data = getXmlData();
    int curCol = 0, index;
    System.err.println("nb:"+data.jjtGetNumChildren());
    for(index=0; index<data.jjtGetNumChildren(); index++) {
      if(data.jjtGetChild(index) instanceof ASTXMLElement &&
         ((ASTXMLElement)data.jjtGetChild(index)).getTagName()!=null) {
	curCol++;
        if(col==curCol) return index;
      }
    }
    if(readColumnFromFile) {
      ASTXMLElement found;
      while((found = parseNextColumn())!=null) {
	index++;
	if(found.getTagName()!=null) {
          curCol++;
	  if(col==curCol) return index;
	}
      }
    }
    throw new ExecuteException("Column "+col+" is out of bound (>"+curCol+")");
  }

}
